from __future__ import division  # TODO: refactor

from decimal import Decimal

from django.conf import settings
from django.core.validators import EMPTY_VALUES, MaxValueValidator, MinValueValidator
from django.db import models, transaction
from django.db.models import JSONField, Max, Sum
from django.urls import reverse
from django.utils.timezone import now
from django.utils.functional import cached_property
from django.utils.module_loading import import_string
from django.utils.translation import gettext_lazy as _

from django_countries.fields import CountryField
from djmoney.forms.widgets import CURRENCY_CHOICES
from internationalflavor.iban.models import IBANField, BICField
from internationalflavor.vat_number import VATNumberField
from model_utils import Choices
from model_utils.fields import MonitorField

from invoicing import settings as invoicing_settings
from invoicing.querysets import InvoiceQuerySet, ItemQuerySet
from invoicing.taxation import TaxationPolicy
from invoicing.taxation.eu import EUTaxationPolicy


def default_supplier(attribute_lookup):
    supplier = getattr(settings, 'INVOICING_SUPPLIER', None)

    if not supplier:
        return None

    lookup_object = supplier
    for attribute in attribute_lookup.split('.'):
        lookup_object = lookup_object.get(attribute, None)

    return lookup_object


class Invoice(models.Model):
    """
    Model representing Invoice itself.
    It keeps all necessary information described at https://www.gov.uk/vat-record-keeping/vat-invoices
    """
    COUNTER_PERIOD = Choices(
        ('DAILY', _('daily')),
        ('MONTHLY', _('monthly')),
        ('YEARLY', _('yearly')),
        ('INFINITE', _('infinite'))
    )

    TYPE = Choices(
        ('INVOICE', _(u'Invoice')),
        ('ADVANCE', _(u'Advance invoice')),
        ('PROFORMA', _(u'Proforma invoice')),
        ('CREDIT_NOTE', _(u'Credit note'))
    )

    STATUS = Choices(
        ('NEW', _(u'new')),
        ('SENT', _(u'sent')),
        ('RETURNED', _(u'returned')),
        ('CANCELED', _(u'canceled')),
        ('PAID', _(u'paid')),
        ('CREDITED', _(u'credited')),
        ('UNCOLLECTIBLE', _(u'uncollectible')),
    )

    PAYMENT_METHOD = Choices(
        ('BANK_TRANSFER', _(u'bank transfer')),
        ('CASH', _(u'cash')),
        ('CASH_ON_DELIVERY', _(u'cash on delivery')),
        ('PAYMENT_CARD', _(u'payment card'))
    )

    DELIVERY_METHOD = Choices(
        ('PERSONAL_PICKUP', _(u'personal pickup')),
        ('MAILING', _(u'mailing')),
        ('DIGITAL', _(u'digital')),
    )

    CONSTANT_SYMBOL = Choices(
        ('0001', _(u'0001 - Payments for goods based on legal and executable decision from legal authority')),
        ('0008', _(u'0008 - Cashless payments for goods')),
        ('0038', _(u'0038 - Cashless funds for wages')),
        ('0058', _(u'0058 - Cashless penalty and delinquency charges')),
        ('0068', _(u'0068 - Transfer of funds for wages and other personal costs')),
        ('0138', _(u'0138 - Cashless deductions at source')),
        ('0168', _(u'0168 - Cashless payments in loans')),
        ('0178', _(u'0178 - Sales from provided services')),
        ('0298', _(u'0298 - Other cashless transfers')),
        ('0304', _(u'0304 - Prior payments for services')),
        ('0308', _(u'0308 - Cashless payments for services')),
        ('0358', _(u'0358 - Payments dedicated to payout through post offices')),
        ('0379', _(u'0379 - Other income, income from postal order')),
        ('0498', _(u'0498 - Payments in loans')),
        ('0558', _(u'0558 - Cashless other financial payments')),
        ('0934', _(u'0934 - Benefits - prior payments')),
        ('0968', _(u'0968 - Other cashless transfers')),
        ('1144', _(u'1144 - Prior payment - advance')),
        ('1148', _(u'1148 - Payment - current advance')),
        ('1744', _(u'1744 - Accounting of tax at income tax of physical body and corporate body')),
        ('1748', _(u'1748 - Income tax of physical body and corporate body based on declared tax year')),
        ('3118', _(u'3118 - Insurance and empl. contrib. to insur. co. and the Labor Office')),
        ('3344', _(u'3344 - Penalty from message - prior')),
        ('3348', _(u'3348 - Penalty from message')),
        ('3354', _(u'3354 - Insurance payments by insurance companies')),
        ('3558', _(u'3558 - Cashless insurance payments by insurance companies')),
        ('8147', _(u'8147 - Payment (posted together with the instruction)'))
    )

    # General information
    type = models.CharField(_(u'type'), max_length=64, choices=TYPE, default=TYPE.INVOICE)
    sequence = models.IntegerField(_(u'sequence'), db_index=True, blank=True)
    number = models.CharField(_(u'number'), max_length=128, blank=True)
    status = models.CharField(_(u'status'), choices=STATUS, max_length=64, default=STATUS.NEW)
    subtitle = models.CharField(_(u'subtitle'), max_length=255, blank=True)
    related_document = models.CharField(_(u'related document'), max_length=100, blank=True)
    related_invoices = models.ManyToManyField(to='self', verbose_name=_(u'related invoices'), blank=True)
    language = models.CharField(_(u'language'), max_length=10, choices=invoicing_settings.LANGUAGES)
    note = models.CharField(_(u'note'), max_length=255, blank=True, default=_(u'Thank you for using our services.'))
    date_issue = models.DateField(_(u'issue date'))
    date_tax_point = models.DateField(_(u'tax point date'), help_text=_(u'time of supply'))
    date_due = models.DateField(_(u'due date'), help_text=_(u'payment till'))
    date_sent = MonitorField(monitor='status', when=[STATUS.SENT], verbose_name=_(u'date sent'),
                             blank=True, null=True, default=None)
    date_paid = MonitorField(monitor='status', when=[STATUS.PAID], verbose_name=_(u'date paid'),
                             blank=True, null=True, default=None)
    date_reminder_sent = models.DateTimeField(verbose_name=_(u'date reminder sent'),
                                              blank=True, null=True, default=None)

    # Payment details
    currency = models.CharField(_(u'currency'), max_length=10, choices=CURRENCY_CHOICES)
    credit = models.DecimalField(_(u'credit'), max_digits=10, decimal_places=2, default=0)
    already_paid = models.DecimalField(_(u'already paid'), max_digits=10, decimal_places=2, default=0,
        validators=[MinValueValidator(0)])

    payment_method = models.CharField(_(u'payment method'), choices=PAYMENT_METHOD, max_length=64)
    constant_symbol = models.CharField(_(u'constant symbol'), max_length=64, choices=CONSTANT_SYMBOL, blank=True)
    variable_symbol = models.PositiveIntegerField(_(u'variable symbol'),
        validators=[MinValueValidator(0), MaxValueValidator(9999999999)],
        blank=True, null=True, default=None)
    specific_symbol = models.PositiveIntegerField(_(u'specific symbol'),
        validators=[MinValueValidator(0), MaxValueValidator(9999999999)],
        blank=True, null=True, default=None)
    reference = models.CharField(_(u'reference'), max_length=140, blank=True)

    bank_name = models.CharField(_(u'bank name'), max_length=255, blank=True)
    bank_street = models.CharField(_(u'bank street and number'), max_length=255, blank=True)
    bank_zip = models.CharField(_(u'bank ZIP'), max_length=255, blank=True)
    bank_city = models.CharField(_(u'bank city'), max_length=255, blank=True)
    bank_country = CountryField(_(u'bank country'), max_length=255, blank=True)
    bank_iban = IBANField(verbose_name=_(u'Account number (IBAN)'), default=None)
    bank_swift_bic = BICField(verbose_name=_(u'Bank SWIFT / BIC'), blank=True)

    # Supplier details
    supplier_name = models.CharField(_(u'supplier name'), max_length=255, default=None)
    supplier_street = models.CharField(_(u'supplier street and number'), max_length=255, blank=True)
    supplier_zip = models.CharField(_(u'supplier ZIP'), max_length=255, blank=True)
    supplier_city = models.CharField(_(u'supplier city'), max_length=255, blank=True)
    supplier_country = CountryField(_(u'supplier country'), default=None)
    supplier_registration_id = models.CharField(_(u'supplier Reg. No.'), max_length=255, blank=True)
    supplier_tax_id = models.CharField(_(u'supplier Tax No.'), max_length=255, blank=True)
    supplier_vat_id = VATNumberField(verbose_name=_(u'supplier VAT No.'), blank=True)
    supplier_additional_info = JSONField(_(u'supplier additional information'),
        blank=True, null=True, default=None)  # for example www or legal matters

    # Contact details
    issuer_name = models.CharField(_(u'issuer name'), max_length=255, blank=True)
    issuer_email = models.EmailField(_(u'issuer email'), blank=True)
    issuer_phone = models.CharField(_(u'issuer phone'), max_length=255, blank=True)

    # Customer details
    customer_name = models.CharField(_(u'customer name'), max_length=255)
    customer_street = models.CharField(_(u'customer street and number'), max_length=255, blank=True)
    customer_zip = models.CharField(_(u'customer ZIP'), max_length=255, blank=True)
    customer_city = models.CharField(_(u'customer city'), max_length=255, blank=True)
    customer_country = CountryField(_(u'customer country'))
    customer_registration_id = models.CharField(_(u'customer Reg. No.'), max_length=255, blank=True)
    customer_tax_id = models.CharField(_(u'customer Tax No.'), max_length=255, blank=True)
    customer_vat_id = VATNumberField(verbose_name=_(u'customer VAT No.'), blank=True)
    customer_additional_info = JSONField(_(u'customer additional information'),
        blank=True, null=True, default=None)
    customer_email = models.EmailField(_(u'customer email'), blank=True)
    customer_phone = models.CharField(_(u'customer phone'), max_length=255, blank=True)

    # Shipping details
    shipping_name = models.CharField(_(u'shipping name'), max_length=255, blank=True)
    shipping_street = models.CharField(_(u'shipping street and number'), max_length=255, blank=True)
    shipping_zip = models.CharField(_(u'shipping ZIP'), max_length=255, blank=True)
    shipping_city = models.CharField(_(u'shipping city'), max_length=255, blank=True)
    shipping_country = CountryField(_(u'shipping country'), blank=True)

    # Delivery details
    delivery_method = models.CharField(_(u'delivery method'), choices=DELIVERY_METHOD, max_length=64,
        default=DELIVERY_METHOD.PERSONAL_PICKUP)

    # sums (auto calculated fields)
    total = models.DecimalField(_(u'total'), max_digits=10, decimal_places=2,
        blank=True, default=0)
    vat = models.DecimalField(_(u'VAT'), max_digits=10, decimal_places=2,
        blank=True, null=True, default=0)

    # Other
    attachments = JSONField(_(u'attachments'),
        blank=True, null=True, default=None)
    created = models.DateTimeField(_(u'created'), auto_now_add=True)
    modified = models.DateTimeField(_(u'modified'), auto_now=True)
    objects = InvoiceQuerySet.as_manager()

    class Meta:
        db_table = 'invoicing_invoices'
        verbose_name = _(u'invoice')
        verbose_name_plural = _(u'invoices')
        ordering = ('date_issue', 'sequence')
        default_permissions = ('list', 'view', 'add', 'change', 'delete')

    def __str__(self):
        return self.number

    def __unicode__(self):
        return self.number

    @transaction.atomic
    def save(self, **kwargs):
        if self.sequence in EMPTY_VALUES:
            self.sequence = Invoice.get_next_sequence(
                type=self.type,
                important_date=self.date_issue,
                number_prefix=getattr(self, 'number_prefix', None),
                generator=getattr(self, 'sequence_generator', None))
        if self.number in EMPTY_VALUES:
            self.number = self._get_number()

        return super(Invoice, self).save(**kwargs)

    def get_absolute_url(self):
        return getattr(settings, 'INVOICING_INVOICE_ABSOLUTE_URL',
            lambda invoice: reverse('invoicing:invoice_detail', args=(invoice.pk,))
        )(self)

    @staticmethod
    def get_next_sequence(type, important_date, number_prefix=None, related_invoices=None, generator=None):
        """
        Returns next invoice sequence based on ``settings.INVOICING_SEQUENCE_GENERATOR``.
        """

        if not generator:
            generator = getattr(settings, 'INVOICING_SEQUENCE_GENERATOR', 'invoicing.helpers.sequence_generator')
            generator = import_string(generator)

        return generator(
            type=type,
            important_date=important_date,
            number_prefix=number_prefix,
            counter_period=None,
            related_invoices=related_invoices
        )

    def _get_number(self):
        """
        Returns next invoice sequence based on ``settings.INVOICING_NUMBER_FORMATTER``.
        """

        if hasattr(self, 'number_formatter'):
            formatter = self.number_formatter
        else:
            formatter = getattr(settings, 'INVOICING_NUMBER_FORMATTER', 'invoicing.helpers.number_formatter')
            formatter = import_string(formatter)

        return formatter(self)

    def get_tax_rate(self):
        if self.taxation_policy:
            # There is taxation policy -> get tax rate
            return self.taxation_policy.get_tax_rate(self)
        else:
            supplier_country_code = self.supplier_country.code if self.supplier_country else None

            # If there is not any special taxation policy, set default tax rate
            return TaxationPolicy.get_default_tax(supplier_country_code, self.date_tax_point)

    @property
    def taxation_policy(self):
        taxation_policy = getattr(settings, 'INVOICING_TAXATION_POLICY', None)

        if taxation_policy is not None:
            return import_string(taxation_policy)

        # Check if supplier is from EU
        if self.supplier_country:
            if EUTaxationPolicy.is_in_EU(self.supplier_country.code):
                return EUTaxationPolicy

        return None

    @property
    def is_overdue(self):
        if self.total == 0:
            return False

        if self.status in [self.STATUS.PAID, self.STATUS.CANCELED, self.STATUS.CREDITED]:
            return False

        if self.type == self.TYPE.CREDIT_NOTE:
            return False

        return self.date_due < now().date()

    @property
    def overdue_days(self):
        return (now().date() - self.date_due).days

    @property
    def days_to_overdue(self):
        return (self.date_due - now().date()).days

    @property
    def payment_term(self):
        return (self.date_due - self.date_issue).days if self.total > 0 else 0

    def set_supplier_data(self, supplier):
        self.supplier_name = supplier.get('name')
        self.supplier_street = supplier.get('street', '')
        self.supplier_zip = supplier.get('zip', '')
        self.supplier_city = supplier.get('city', '')
        self.supplier_country = supplier.get('country_code')
        self.supplier_registration_id = supplier.get('registration_id', '')
        self.supplier_tax_id = supplier.get('tax_id', '')
        self.supplier_vat_id = supplier.get('vat_id', '')
        self.supplier_additional_info = supplier.get('additional_info', None)

        bank = supplier.get('bank')
        self.bank_name = bank.get('name')
        self.bank_street = bank.get('street')
        self.bank_zip = bank.get('zip')
        self.bank_city = bank.get('city')
        self.bank_country = bank.get('country_code')
        self.bank_iban = bank.get('iban')
        self.bank_swift_bic = bank.get('swift_bic')

    def set_customer_data(self, customer):
        self.customer_name = customer.get('name')
        self.customer_street = customer.get('street', '')
        self.customer_zip = customer.get('zip', '')
        self.customer_city = customer.get('city', '')
        self.customer_country = customer.get('country_code')
        self.customer_registration_id = customer.get('registration_id', '')
        self.customer_tax_id = customer.get('tax_id', '')
        self.customer_vat_id = customer.get('vat_id', '')
        self.customer_additional_info = customer.get('additional_info', None)

    def set_shipping_data(self, shipping):
        self.shipping_name = shipping.get('name', '')
        self.shipping_street = shipping.get('street', '')
        self.shipping_zip = shipping.get('zip', '')
        self.shipping_city = shipping.get('city', '')
        self.shipping_country = shipping.get('country_code', '')

    # http://www.superfaktura.sk/blog/neplatca-dph-vzor-faktury/
    def is_supplier_vat_id_visible(self):
        # TODO: deprecated method
        is_supplier_vat_id_visible = getattr(settings, 'INVOICING_IS_SUPPLIER_VAT_ID_VISIBLE', None)

        if is_supplier_vat_id_visible is not None:
            return is_supplier_vat_id_visible(self)

        if self.vat is None and self.supplier_country == self.customer_country:
            return False

        # VAT is not 0
        if self.vat != 0 or self.item_set.filter(tax_rate__gt=0).exists():
            return True

        # VAT is 0, check if customer is from EU and from same country as supplier
        return self.is_EU_customer() and self.supplier_country != self.customer_country

    def is_EU_customer(self):
        return EUTaxationPolicy.is_in_EU(self.customer_country.code) if self.customer_country else False

    def is_reverse_charge(self):
        return self.taxation_policy.is_reverse_charge(self)

    @property
    def vat_summary(self):
        # rates_and_sum = self.item_set.all().annotate(base=Sum(F('qty')*F('price_per_unit'))).values('tax_rate', 'base')
        # rates_and_sum = self.item_set.all().values('tax_rate').annotate(Sum('price_per_unit'))
        # rates_and_sum = self.item_set.all().values('tax_rate').annotate(Sum(F('qty')*F('price_per_unit')))

        from django.db import connection
        cursor = connection.cursor()
        cursor.execute('select tax_rate as rate, SUM(quantity*unit_price*(100-discount)/100) as base, ROUND(CAST(SUM(quantity*unit_price*((100-discount)/100)*(tax_rate/100)) AS numeric), 2) as vat from invoicing_items where invoice_id = %s group by tax_rate;', [self.pk])

        desc = cursor.description
        return [
            dict(zip([col[0] for col in desc], row))
            for row in cursor.fetchall()
        ]

    @cached_property
    def has_discount(self):
        if not self.item_set.exists():
            return False

        discounts = list(set(self.item_set.values_list('discount', flat=True)))
        return len(discounts) > 1 or discounts[0] > 0

    @cached_property
    def has_unit(self):
        if not self.item_set.exists():
            return False

        units = list(set(self.item_set.values_list('unit', flat=True)))
        return len(units) > 1 or units[0] != Item.UNIT_EMPTY

    @cached_property
    def max_quantity(self):
        quantity = self.item_set.aggregate(Max('quantity'))
        return quantity.get('quantity__max', 1) if quantity else 0

    @cached_property
    def sum_quantity(self):
        quantity = self.item_set.aggregate(Sum('quantity'))
        return quantity.get('quantity__sum', 1) if quantity else 0

    @cached_property
    def all_items_with_single_quantity(self):
        return self.item_set.count() == self.sum_quantity

    @property
    def subtotal(self):
        sum = 0
        for item in self.item_set.all():
            sum += item.subtotal

        sum -= Decimal(self.credit)  # subtract credit

        return round(sum, 2)

    @property
    def discount(self):
        sum = 0
        for item in self.item_set.all():
            sum += item.discount_amount
        return round(sum, 2)

    @property
    def discount_percentage(self):
        percentage = 100 * self.discount / self.total_without_discount
        return round(percentage, 2)

    @property
    def total_without_discount(self):
        return Decimal(self.total) + self.discount + Decimal(self.credit)

    @property
    def to_pay(self):
        return self.total - self.already_paid

    def calculate_vat(self):
        if len(self.vat_summary) == 1 and self.vat_summary[0]['vat'] is None:
            return None

        vat = 0
        for vat_rate in self.vat_summary:
            vat += vat_rate['vat'] or 0
        return vat

    def calculate_total(self):
        total = 0

        for vat_rate in self.vat_summary:
            total += Decimal(vat_rate['base']) + Decimal(vat_rate['vat'] or 0)

        total -= Decimal(self.credit)  # subtract credit
        return round(total, 2)

    def create_copy(self, **kwargs):
        # prepare new instance data
        from django.forms import model_to_dict
        invoice_dict = model_to_dict(self, exclude=['id', 'related_invoices', 'number', 'sequence'])

        # update data with custom params
        invoice_dict.update(kwargs)

        # check presence of custom sequence generator and number formatter
        sequence_generator = invoice_dict.pop('sequence_generator')
        number_formatter = invoice_dict.pop('number_formatter')

        # create new instance but don't save yet
        new_instance = Invoice(**invoice_dict)

        # pass custom sequence generator and number formatter
        new_instance.sequence_generator = sequence_generator
        new_instance.number_formatter = number_formatter

        # save new instance with new sequence and number
        new_instance.save()

        # set current invoice as related invoice
        new_instance.related_invoices.set([self])

        # duplicate items
        for item in self.item_set.all():
            item_kwargs = model_to_dict(item, exclude=['id', 'invoice'])
            item_kwargs.update({'invoice': new_instance})
            Item.objects.create(**item_kwargs)

        # return copied invoice
        return new_instance


class Item(models.Model):
    WEIGHT = [(i, i) for i in range(0, 20)]
    UNIT_EMPTY = 'EMPTY'
    UNIT_PIECES = 'PIECES'
    UNIT_HOURS = 'HOURS'
    UNITS = (
        (UNIT_EMPTY, ''),
        (UNIT_PIECES, _(u'pcs.')),
        (UNIT_HOURS, _(u'hours'))
    )

    invoice = models.ForeignKey(Invoice, verbose_name=_(u'invoice'), on_delete=models.CASCADE)
    title = models.TextField(_(u'title'))
    quantity = models.DecimalField(_(u'quantity'), max_digits=10, decimal_places=3, default=1)
    unit = models.CharField(_(u'unit'), choices=UNITS, max_length=64, default=UNIT_PIECES)
    unit_price = models.DecimalField(_(u'unit price'), max_digits=10, decimal_places=2)
    discount = models.DecimalField(_(u'discount (%)'), max_digits=4, decimal_places=1, default=0)
    tax_rate = models.DecimalField(_(u'tax rate (%)'), max_digits=3, decimal_places=1,
        help_text=_(u'Enter amount or keep blank for no tax or reverse charge (if applicable)'),
        blank=True, null=True, default=None)
    tag = models.CharField(_(u'tag'), max_length=128,
        blank=True, null=True, default=None)
    weight = models.IntegerField(_(u'weight'), choices=WEIGHT, help_text=_(u'ordering'),
        blank=True, null=True, default=0)
    created = models.DateTimeField(_(u'created'), auto_now_add=True)
    modified = models.DateTimeField(_(u'modified'), auto_now=True)
    objects = ItemQuerySet.as_manager()

    class Meta:
        db_table = 'invoicing_items'
        verbose_name = _(u'item')
        verbose_name_plural = _(u'items')
        ordering = ('-invoice', 'weight', 'created')

    def __str__(self):
        return self.title

    def get_absolute_url(self):
        return getattr(settings, 'INVOICING_INVOICE_ITEM_ABSOLUTE_URL', lambda item: '')(self)

    @property
    def subtotal(self):
        subtotal = round(self.unit_price * self.quantity, 2)
        return round(Decimal(subtotal) * Decimal((100 - self.discount) / 100), 2)

    @property
    def discount_amount(self):
        subtotal = round(self.unit_price_with_vat * self.quantity, 2)
        return round(Decimal(subtotal) * Decimal(self.discount / 100), 2)

    @property
    def vat(self):
        return round(self.subtotal * Decimal(self.tax_rate) / 100 if self.tax_rate else 0, 2)

    @property
    def unit_price_with_vat(self):
        tax_rate = self.tax_rate if self.tax_rate else 0
        return round(Decimal(self.unit_price) * Decimal((100 + tax_rate) / 100), 2)

    @property
    def total(self):
        return round(self.subtotal + self.vat, 2)

    def save(self, **kwargs):
        # TODO: move to validator
        if self.tax_rate not in EMPTY_VALUES and self.invoice.supplier_vat_id in EMPTY_VALUES:
            raise ValueError(f'Tax rate is {self.tax_rate}% but supplier VAT ID is not set. Invoice #{self.invoice.pk}, number {self.invoice.number}')

        # TODO: find out if user explicitly set None as value or should be set automatically
        # if self.tax_rate in EMPTY_VALUES and self.pk is None:
        # If tax rate is not set while creating new invoice item, set it according billing details
        # self.tax_rate = self.invoice.get_tax_rate()

        return super(Item, self).save(**kwargs)


from .signals import *



import binascii

from django.conf import settings
from django.utils.module_loading import import_string
from requests_futures import sessions


def get_invoices_in_pdf(invoices):
    # TODO: replace with invoicing_settings
    invoicing_formatter = getattr(settings, 'INVOICING_FORMATTER', 'invoicing.formatters.html.BootstrapHTMLFormatter')
    formatter_class = import_string(invoicing_formatter)
    htmltopdf_api_url = getattr(settings, 'HTMLTOPDF_API_URL', None)
    printmyweb_url = getattr(settings, 'PRINTMYWEB_URL', None)
    printmyweb_token = getattr(settings, 'PRINTMYWEB_TOKEN', None)
    requests = []
    export_files = []

    print_api_url = htmltopdf_api_url or printmyweb_url

    for invoice in invoices:
        formatter = formatter_class(invoice)
        invoice_content = formatter.get_response().content

        hex_4_bytes = binascii.hexlify(invoice_content)[0:8]

        # Look at the first 4 bytes of the file.
        # PDF has "%PDF" (hex 25 50 44 46) and ZIP has hex 50 4B 03 04.
        is_pdf = hex_4_bytes == b'25504446'

        if is_pdf:
            export_files.append({'name': str(invoice) + '.pdf', 'content': invoice_content})
        else:
            if print_api_url is None:
                raise NotImplementedError('Invoice content is not PDF and print_api_url is not set!')
            else:
                requests.append({'invoice': str(invoice), 'html_content': invoice_content})

    if len(requests) > 0:
        session = sessions.FuturesSession(max_workers=3)

        kwargs = {}
        if printmyweb_token:
            kwargs['headers'] = {
                'api-key': printmyweb_token
            }

        futures = [
            {'invoice': request.get('invoice'), 'future': session.post(
                url=print_api_url,
                data=request.get('html_content'),
                **kwargs
            )}
            for request in requests]

        for f in futures:
            file_name = f.get('invoice')
            result = f.get('future').result()
            export_files.append({'name': file_name + '.pdf', 'content': result.content})

    return export_files



# Generated by Django 4.0.8 on 2023-12-13 16:41

from django.db import migrations, models


def echo(*args, **kwargs):
    print('This migration will be deleted in the next release. If you already applied previous version, drop supplier_is_vat_payer field manually!')

class Migration(migrations.Migration):

    dependencies = [
        ('invoicing', '0029_auto_20230912_0458'),
    ]

    operations = [
        # TODO: DEPRECATED: previous version of migration adding new field
        # migrations.AddField(
        #     model_name='invoice',
        #     name='supplier_is_vat_payer',
        #     field=models.BooleanField(blank=True, null=True, default=None)
        # ),
        migrations.RunPython(echo, migrations.RunPython.noop)
    ]



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.core.validators import EMPTY_VALUES
from django.db import transaction
from django.db.models import Max
from django.template import Template, Context
from django.utils.translation import gettext_lazy as _

from invoicing.models import Invoice


def sequence_generator(type, important_date, number_prefix=None, counter_period=None, related_invoices=None, start_from=None):
    """
    Returns next invoice sequence based on ``settings.INVOICING_COUNTER_PERIOD``.

    .. warning::

        This is only used to prepopulate ``sequence`` field on saving new invoice.
        To get invoice sequence always use ``sequence`` field.

    .. note::

        To get invoice number use ``number`` field.

    :return: string (generated next sequence)
    """
    with transaction.atomic():
        Invoice.objects.lock()

        if not counter_period:
            counter_period = getattr(settings, 'INVOICING_COUNTER_PERIOD', Invoice.COUNTER_PERIOD.YEARLY)

        if related_invoices is None:
            related_invoices = Invoice.objects.all()

        if counter_period == Invoice.COUNTER_PERIOD.DAILY:
            related_invoices = related_invoices.filter(date_issue=important_date)

        elif counter_period == Invoice.COUNTER_PERIOD.YEARLY:
            related_invoices = related_invoices.filter(date_issue__year=important_date.year)

        elif counter_period == Invoice.COUNTER_PERIOD.MONTHLY:
            related_invoices = related_invoices.filter(date_issue__year=important_date.year, date_issue__month=important_date.month)

        elif counter_period != Invoice.COUNTER_PERIOD.INFINITE:
            raise ImproperlyConfigured("INVOICING_COUNTER_PERIOD can be set only to these values: DAILY, MONTHLY, YEARLY, INFINITE.")

        invoice_counter_per_type = getattr(settings, 'INVOICING_COUNTER_PER_TYPE', False)

        if invoice_counter_per_type:
            if type in EMPTY_VALUES:
                raise ValueError(_('Invoice type is required when INVOICING_COUNTER_PER_TYPE is enabled'))

            related_invoices = related_invoices.filter(type=type)
        elif type not in EMPTY_VALUES:
            # TODO: log instead
            print(_('Invoice type specified but INVOICING_COUNTER_PER_TYPE is disabled'))

        if number_prefix is not None:
            related_invoices = related_invoices.filter(number__startswith=number_prefix)

        start_from = start_from if start_from is not None else getattr(settings, 'INVOICING_NUMBER_START_FROM', 1)
        last_sequence = related_invoices.aggregate(Max('sequence'))['sequence__max'] or start_from - 1

        return last_sequence + 1


def number_formatter(invoice):
    """
    Generates on the fly invoice number from template provided by ``settings.INVOICING_NUMBER_FORMAT``.
    ``Invoice`` object is provided as ``invoice`` variable to the template, therefore all object fields
    can be used to generate number format.

    .. warning::

        This is only used to prepopulate ``number`` field on saving new invoice.
        To get invoice number always use ``number`` field.

    :return: string (generated number)
    """

    # default settings number format
    default_number_format = getattr(settings, "INVOICING_NUMBER_FORMAT", "{{ invoice.date_issue|date:'Y' }}/{{ invoice.sequence }}")
    # specific invoice number format
    number_format = getattr(invoice, 'number_format', default_number_format)

    # render number by given format
    return Template(number_format).render(Context({'invoice': invoice}))



from django.apps import AppConfig
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.timezone import now
from dateutil.relativedelta import relativedelta

try:
    # older Django
    from django.utils.translation import ugettext_lazy as _
except ImportError:
    # Django >= 3
    from django.utils.translation import gettext_lazy as _


class Config(AppConfig):
    name = 'invoicing'
    verbose_name = _('Invoicing')

    def ready(self):
        from invoicing.models import Invoice

        counter_period = getattr(settings, "INVOICING_COUNTER_PERIOD")
        sequence = 1
        date = now()
        invoice1 = Invoice(date_issue=date, date_tax_point=date, sequence=sequence)

        if counter_period == Invoice.COUNTER_PERIOD.DAILY:
            date += relativedelta(days=1)
        elif counter_period == Invoice.COUNTER_PERIOD.MONTHLY:
            date += relativedelta(months=1)
        elif counter_period == Invoice.COUNTER_PERIOD.YEARLY:
            date += relativedelta(years=1)
        elif counter_period == Invoice.COUNTER_PERIOD.INFINITE:
            sequence += 1

        invoice2 = Invoice(date_issue=date, date_tax_point=date, sequence=sequence)

        if invoice1._get_number() == invoice2._get_number():
            raise ImproperlyConfigured("The INVOICING_NUMBER_FORMAT is incorrect for the current INVOICING_COUNTER_PERIOD")



# Generated by Django 2.2.23 on 2023-09-12 02:58

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('invoicing', '0028_auto_20230912_0404'),
    ]

    operations = [
        migrations.AlterField(
            model_name='invoice',
            name='already_paid',
            field=models.DecimalField(decimal_places=2, default=0, max_digits=10, validators=[django.core.validators.MinValueValidator(0)]),
        ),
    ]



import json

import requests
from django.utils import translation
from django.utils.module_loading import import_string
from django.utils.safestring import mark_safe
from django_filters.constants import EMPTY_VALUES

from django.utils.translation import gettext_lazy as _

from invoicing import settings as invoicing_settings
from invoicing.models import Invoice


def get_accounting_software_manager():
    if invoicing_settings.ACCOUNTING_SOFTWARE_MANAGER is not None:
        return import_string(invoicing_settings.ACCOUNTING_SOFTWARE_MANAGER)()

    if invoicing_settings.ACCOUNTING_SOFTWARE not in EMPTY_VALUES:
        # TODO: use title()
        if invoicing_settings.ACCOUNTING_SOFTWARE == 'IKROS':
            return IKrosManager()

        if invoicing_settings.ACCOUNTING_SOFTWARE == 'PROFIT365':
            return Profit365Manager()

        return NotImplementedError(_('Accounting software %s not implemented') % invoicing_settings.ACCOUNTING_SOFTWARE)

    return None


class AccountingSoftwareManager(object):
    def send_to_accounting_software(self, request, queryset):
        raise NotImplementedError()


class IKrosManager(AccountingSoftwareManager):
    def __init__(self):
        if invoicing_settings.ACCOUNTING_SOFTWARE_API_DATA in EMPTY_VALUES:
            raise EnvironmentError(_('Missing accounting software API key'))

    def send_to_accounting_software(self, request, queryset):
        invoices_data = []

        for invoice in queryset:
            invoice_data = {
                "documentNumber": invoice.number,
                "createDate": invoice.date_issue.strftime('%Y-%m-%dT00:00:00'),
                "dueDate": invoice.date_due.strftime('%Y-%m-%dT00:00:00'),
                "completionDate": invoice.date_tax_point.strftime('%Y-%m-%dT00:00:00'),
                "clientName": invoice.customer_name,
                "clientStreet": invoice.customer_street,
                "clientPostCode": invoice.customer_zip,
                "clientTown": invoice.customer_city,
                "clientCountry": invoice.get_customer_country_display(),
                "clientRegistrationId": invoice.customer_registration_id,
                "clientTaxId": invoice.customer_tax_id,
                "clientVatId": invoice.customer_vat_id,
                "clientPhone": invoice.customer_phone,
                "clientEmail": invoice.customer_email,
                "variableSymbol": invoice.variable_symbol,
                "paymentType": invoice.get_payment_method_display(),
                "deliveryType": invoice.get_delivery_method_display(),
                "senderContactName": invoice.issuer_name,
                "clientPostalName": invoice.shipping_name,
                "clientPostalStreet": invoice.shipping_street,
                "clientPostalPostCode": invoice.shipping_zip,
                "clientPostalTown": invoice.shipping_city,
                "clientPostalCountry": invoice.get_shipping_country_display(),
                "clientInternalId": f'{invoice.customer_country}001',  # TODO: custom mapping
                # "clientHasDifferentPostalAddress": True,
                "currency": invoice.currency,
                # "orderNumber": invoice.variable_symbol,
                "items": []
            }

            for item in invoice.item_set.all():
                item_data = {
                    "name": item.title,
                    "count": str(item.quantity),
                    "measureType": item.get_unit_display(),
                    "unitPrice": str(item.unit_price),
                    "vat": item.vat
                }

                if invoice.status == Invoice.STATUS.CANCELED:
                    item_data['count'] = 0  # not working actually (min value = 1)
                    item_data['unitPrice'] = 0
                    # item_data['description'] = 'STORNO'

                invoice_data['items'].append(item_data)

            if invoice.status == Invoice.STATUS.CANCELED:
                invoice_data['closingText'] = 'STORNO'

            if invoice.credit != 0:
                invoice_data['items'][0]['hasDiscount'] = True
                invoice_data['items'][0]['discountValue'] = str(invoice.credit * -1)  # TODO: substract VAT
                # invoice_data['items'][0]['discountValueWithVat'] = str(invoice.credit * -1)

            invoices_data.append(invoice_data)

        # pprint(invoices_data)  # TODO: logging
        payload = json.dumps(invoices_data)

        url = invoicing_settings.ACCOUNTING_SOFTWARE_IKROS_API_URL
        api_key = invoicing_settings.ACCOUNTING_SOFTWARE_API_DATA['apiKey']
        headers = {
            'Authorization': 'Bearer ' + str(api_key),
            'Content-Type': 'application/json'
        }
        r = requests.post(url=url, data=payload, headers=headers)
        data = r.json()

        # pprint(data)  # TODO: logging

        if data.get('message', None) is not None:
            raise Exception(_('Result code: %d. Message: %s (%s)') % (
                data.get('code', data.get('resultCode')),
                data['message'],
                data.get('errorType', '-')
            ))

        if 'documents' in data:
            if len(data['documents']) > 0:
                download_url = data['documents'][0]['downloadUrl']
                # requests.get(download_url)

                return mark_safe(_('%d invoices sent to accounting software [<a href="%s" target="_blank">Fetch</a>]') % (
                    queryset.count(),
                    download_url
                ))
            else:
                return _('%d invoices sent to accounting software') % (
                    queryset.count(),
                )


class Profit365Manager(AccountingSoftwareManager):
    def __init__(self):
        if invoicing_settings.ACCOUNTING_SOFTWARE_API_DATA in EMPTY_VALUES:
            raise EnvironmentError(_('Missing accounting software API data'))

    def send_to_accounting_software(self, request, queryset):
        results = []

        for invoice in queryset:
            with translation.override(invoice.language):
                partnerAddress = [
                    invoice.customer_name,
                    invoice.customer_street,
                    invoice.customer_zip,
                    invoice.customer_city,
                    invoice.get_customer_country_display()
                ]

                if invoice.customer_registration_id:
                    partnerAddress += [
                        '%s: %s' % (_('Reg. No.'), invoice.customer_registration_id),
                    ]

                if invoice.customer_tax_id:
                    partnerAddress += [
                        '%s: %s' % (_('Tax No.'), invoice.customer_tax_id),
                    ]

                if invoice.customer_vat_id:
                    partnerAddress += [
                        '%s: %s' % (_('VAT No.'), invoice.customer_vat_id),
                    ]

                # "clientEmail": invoice.customer_email,  # partner email

                invoice_data = {
                    "recordNumber": invoice.number,
                    "bankAccountId": invoicing_settings.ACCOUNTING_SOFTWARE_API_DATA['bankAccountId'],
                    # "ordnerID": "VYF",  # TODO: setting
                    # "warehouseID": "TODO", # TODO: setting
                    # "partnerDetail": "Firma s.r.o.\r\nLesná 123\r\nBratislava\r\nIČO: test",
                    "partnerAddress": '\r\n'.join(partnerAddress),  # delivery address
                    # "partnerId": "TODO",  # TODO: find partner by externalId
                    "phone": invoice.customer_phone,
                    "email": invoice.customer_email,
                    "dateCreated": invoice.date_issue.strftime('%Y-%m-%dT00:00:00'),
                    "dateAccounting": invoice.date_tax_point.strftime('%Y-%m-%dT00:00:00'),
                    # "dateDelivery": "2020-07-20",
                    "dateValidTo": invoice.date_due.strftime('%Y-%m-%dT00:00:00'),  # due date
                    # "orderRecordNo": invoice.variable_symbol,  # TODO: order number
                    "symbolSpecific": invoice.specific_symbol,
                    "symbolVariable": invoice.variable_symbol,
                    "symbolConstant": invoice.constant_symbol,
                    "localCurrencyID": invoice.currency,
                    "currencyID": invoice.currency,
                    "issuedBy": invoice.issuer_name,
                    # "bonusPercent": invoice.credit, # TODO: calculate
                    # "deliveryTypeId": postou  # TODO: invoice.delivery_method
                    # "paymentTypeId": prevod / kartou  # TODO: invoice.payment_method

                    # "paymentType": invoice.get_payment_method_display(),
                    # "deliveryType": invoice.get_delivery_method_display(),

                    # "clientPostalName": invoice.shipping_name,
                    # "clientPostalStreet": invoice.shipping_street,
                    # "clientPostalPostCode": invoice.shipping_zip,
                    # "clientPostalTown": invoice.shipping_city,
                    # "clientPostalCountry": invoice.get_shipping_country_display(),
                    # "clientInternalId": f'{invoice.customer_country}001',  # TODO: partner: external Id
                    # "clientHasDifferentPostalAddress": True,
                    "rows": []
                }

                if invoice.status == Invoice.STATUS.CANCELED:
                    invoice_data['description'] = 'STORNO'
                    invoice_data['commentBelowItems'] = 'STORNO 2'
                else:
                    for item in invoice.item_set.all():
                        item_data = {
                            # "itemId": "808497a7-c5cb-4d7e-a3d2-6c7e49735831",  # TODO: find by external Id
                            "name": item.title,
                            "price": str(item.unit_price),
                            "quantity": str(item.quantity)
                            # "measureType": item.get_unit_display(),
                            # "vat": item.vat
                        }

                        if item.discount > 0:
                            item_data['discountPercent'] = str(item.discount)

                        # if invoice.status == Invoice.STATUS.CANCELED:
                        #     item_data['quantity'] = 0  # not working actually (min value = 1)
                        #     item_data['price'] = 0
                        #     invoice_data['description'] = 'STORNO 1'
                        #     invoice_data['commentBelowItems'] = 'STORNO 2'

                        invoice_data['rows'].append(item_data)

                # if invoice.status == Invoice.STATUS.CANCELED:
                #     invoice_data['closingText'] = 'STORNO'

                # if invoice.credit != 0:
                #     invoice_data['items'][0]['hasDiscount'] = True
                #     invoice_data['items'][0]['discountValue'] = str(invoice.credit * -1)  # TODO: substract VAT
                    ## invoice_data['items'][0]['discountValueWithVat'] = str(invoice.credit * -1)

            # pprint(invoice_data)  # TODO: logging
            payload = json.dumps(invoice_data)

            url = invoicing_settings.ACCOUNTING_SOFTWARE_PROFIT365_API_URL
            headers = {
                'Authorization': invoicing_settings.ACCOUNTING_SOFTWARE_API_DATA['Authorization'],
                'ClientID': invoicing_settings.ACCOUNTING_SOFTWARE_API_DATA['ClientID'],
                'ClientSecret': invoicing_settings.ACCOUNTING_SOFTWARE_API_DATA['ClientSecret'],
                'Content-Type': 'application/json'
            }
            if 'CompanyID' in invoicing_settings.ACCOUNTING_SOFTWARE_API_DATA:
                headers['CompanyID'] = invoicing_settings.ACCOUNTING_SOFTWARE_API_DATA

            r = requests.post(url=f'{url}/sales/invoices', data=payload, headers=headers)

            # if r.status_code == 200:
            #     data = r.json()
                # pprint(data)

            results.append({
                'invoice': invoice.number,
                'status_code': r.status_code,
                'reason': r.reason
            })

        return results



from django.db.models.signals import post_save, post_delete, pre_save
from django.dispatch import receiver
from invoicing.models import Item, Invoice
# from pragmatic.signals import temporary_disconnect_signal


@receiver(post_save, sender=Item)
@receiver(post_delete, sender=Item)
def recalculate_total_by_items(instance, **kwargs):
    invoice = instance.invoice
    invoice.total = invoice.calculate_total()
    invoice.vat = invoice.calculate_vat()
    # with temporary_disconnect_signal(signal=post_save, receiver=recalculate_total_by_invoice, sender=Invoice):
    invoice.save(update_fields=['total', 'vat'])


@receiver(pre_save, sender=Invoice)
def recalculate_total_by_invoice(instance, **kwargs):
    invoice = instance
    invoice.total = invoice.calculate_total()
    invoice.vat = invoice.calculate_vat()



from django import template
from django.template import loader, Context

from ..formatters.html import HTMLFormatter

register = template.Library()


@register.filter
def as_html(invoice):
    template = loader.get_template('invoicing/formatters/html.html')
    formatter = HTMLFormatter(invoice)
    data = formatter.get_data()

    try:
        return template.render(Context(data))
    except TypeError:
        return template.render(data)


@register.filter
def nice_iban(iban):
    return ' '.join(iban[i:i+4] for i in range(0, len(iban), 4))



# Generated by Django 2.2.9 on 2020-09-01 10:41

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('invoicing', '0023_invoice_related_invoices'),
    ]

    operations = [
        migrations.AddField(
            model_name='invoice',
            name='date_reminder_sent',
            field=models.DateTimeField(blank=True, default=None, null=True),
        ),
    ]




from django.db import migrations, transaction
from invoicing.models import Invoice


def update_credit_notes(*args, **kwargs):
    Invoice.objects.filter(type='VAT_CREDIT_NOTE').update(type='CREDIT_NOTE')


class Migration(migrations.Migration):
    dependencies = [
        ('invoicing', '0015_auto_20180605_1232'),
    ]

    operations = [
        migrations.RunPython(update_credit_notes, lambda *args, **kwargs: None)
    ]



# Generated by Django 2.2.23 on 2023-09-12 02:04

from django.db import migrations, models
import internationalflavor.vat_number.models


class Migration(migrations.Migration):

    dependencies = [
        ('invoicing', '0027_auto_20230207_2029'),
    ]

    operations = [
        migrations.AddField(
            model_name='invoice',
            name='already_paid',
            field=models.DecimalField(decimal_places=2, default=0, max_digits=10),
        ),
        migrations.AlterField(
            model_name='invoice',
            name='customer_vat_id',
            field=internationalflavor.vat_number.models.VATNumberField(blank=True),
        ),
        migrations.AlterField(
            model_name='invoice',
            name='related_invoices',
            field=models.ManyToManyField(blank=True, related_name='_invoice_related_invoices_+', to='invoicing.Invoice'),
        ),
        migrations.AlterField(
            model_name='invoice',
            name='supplier_vat_id',
            field=internationalflavor.vat_number.models.VATNumberField(blank=True),
        ),
    ]



from collections import OrderedDict

from invoicing.models import Invoice
from invoicing.utils import get_invoices_in_pdf

from django.utils.translation import gettext_lazy as _, gettext

from outputs.mixins import ExporterMixin, ExcelExporterMixin, FilterExporterMixin
from outputs.models import Export
from pragmatic.utils import compress


# TODO: inherit from filterexporter mixin?
# class InvoiceXlsxListExporter(FilterExporterMixin, ExcelExporterMixin):
class InvoiceXlsxListExporter(ExcelExporterMixin):
    # filter_class = InvoiceFilter
    model = Invoice
    queryset = None
    filename = _('invoices.xlsx')

    @classmethod
    def get_model(cls):
        return cls.queryset.model if cls.queryset is not None else cls.model

    @classmethod
    def get_app_and_model(cls):
        return cls.get_model()._meta.label.split('.')

    @staticmethod
    def selectable_fields():
        # attribute, label, width, format (self.FORMATS), value
        return OrderedDict({
            gettext('Details'): [
                ('id', gettext('ID'), 7, 'integer'),
                ('created', gettext('Created'), 20, 'datetime'),
                ('get_type_display', gettext('Type'), 10, None, lambda value: value()),
                ('sequence', gettext('Sequence'), 10),
                ('number', gettext('Number'), 15),
                ('get_status_display', gettext('Status'), 10, None, lambda value: value()),
                ('subtitle', gettext('Subtitle'), 20),
                ('get_language_display', gettext('Language'), 10, None, lambda value: value()),
                ('note', gettext('Note'), 30),
            ],
            gettext('Dates'): [
                ('date_issue', gettext('Issue date'), 15, 'date'),
                ('date_tax_point', gettext('Tax point date'), 15, 'date'),
                ('date_due', gettext('Due date'), 15, 'date'),
                ('date_sent', gettext('Sent (date)'), 15, 'date'),
                ('date_paid', gettext('Date of payment'), 15, 'date', lambda value: value.date() if value else None),
                ('payment_term', gettext('Payment term (days)'), 15, 'integer'),
                ('overdue_days', gettext('Overdue (days)'), 15, 'integer', lambda value, obj: (
                    '=IF(TODAY()<=%(date_due)d, "", _xlfn.DAYS(TODAY(),%(date_due)d))' % {
                        'date_due': ExcelExporterMixin.to_excel_datetime(obj.date_due)
                    } if obj.status not in [Invoice.STATUS.PAID, Invoice.STATUS.CANCELED] else '', obj.overdue_days if obj.is_overdue else '')),
            ],
            gettext('Payment'): [
                ('total', gettext('Total'), 10),
                ('vat', gettext('VAT'), 10),
                ('get_currency_display', gettext('Currency'), 10, None, lambda value: value()),
                ('credit', gettext('Credit'), 10),
                ('get_payment_method_display', gettext('Payment method'), 20, None, lambda value: value()),
                ('constant_symbol', gettext('Constant symbol'), 20),
                ('variable_symbol', gettext('Variable symbol'), 20),
                ('specific_symbol', gettext('Specific symbol'), 20),
                ('reference', gettext('Reference'), 20),
                ('bank_name', gettext('Bank name'), 20),
                # ('bank_street', gettext('Bank street'), 20),
                # ('bank_zip', gettext('Bank zip'), 15),
                # ('bank_city', gettext('Bank city'), 20),
                # ('get_bank_country_display', gettext('Bank country'), 20, None, lambda value: value()),
                ('bank_iban', gettext('IBAN'), 30),
                ('bank_swift_bic', gettext('SWIFT/BIC'), 15),
            ],
            gettext('Issuer'): [
                ('supplier_name', gettext('Supplier name'), 20),
                ('supplier_street', gettext('Supplier street'), 20),
                ('supplier_zip', gettext('Supplier zip'), 15),
                ('supplier_city', gettext('Supplier city'), 15),
                ('get_supplier_country_display', gettext('Supplier country'), 20, None, lambda value: value()),
                ('supplier_registration_id', gettext('Supplier reg. ID'), 20),
                ('supplier_tax_id', gettext('Supplier tax ID'), 20),
                ('supplier_vat_id', gettext('Supplier VAT ID'), 20),
                ('supplier_additional_info', gettext('Supplier additional info'), 25, None, lambda ord_dict: ', '.join(': '.join([str(label), str(value)]) for label, value in ord_dict.items()) if ord_dict and not isinstance(ord_dict, str) else ''),
                ('issuer_name', gettext('Issuer name'), 20),
                ('issuer_email', gettext('Issuer email'), 30),
                ('issuer_phone', gettext('Issuer phone'), 30),
            ],
            gettext('Customer'): [
                ('customer_name', gettext('Customer name'), 20),
                ('customer_street', gettext('Customer street'), 20),
                ('customer_zip', gettext('Customer zip'), 15),
                ('customer_city', gettext('Customer city'), 20),
                ('get_customer_country_display', gettext('Customer country'), 20, None, lambda value: value()),
                ('customer_registration_id', gettext('Customer reg. ID'), 20),
                ('customer_tax_id', gettext('Customer tax ID'), 20),
                ('customer_vat_id', gettext('Customer VAT ID'), 20),
                ('customer_additional_info', gettext('Customer additional info'), 25, None, lambda ord_dict: ', '.join(': '.join([str(label), str(value)]) for label, value in ord_dict.items()) if ord_dict and not isinstance(ord_dict, str) else ''),
                ('customer_email', gettext('Customer email'), 30),
                ('customer_phone', gettext('Customer phone'), 30),
            ],
            # gettext('Shipping'): [
            #     ('shipping_name', gettext('Shipping name'), 20),
            #     ('shipping_street', gettext('Shipping street'), 20),
            #     ('shipping_zip', gettext('Shipping zip'), 15),
            #     ('shipping_city', gettext('Shipping city'), 20),
            #     ('get_shipping_country_display', gettext('Shipping country'), 20, None, lambda value: value()),
            #     ('get_delivery_method_display', gettext('Delivery method'), 20, None, lambda value: value()),
            # ],
        })

    # def get_whole_queryset(self, params):
    #     return super().get_whole_queryset(params) \
    #         .order_by('-created').distinct()
    #         # .prefetch_related(Prefetch('item_set', queryset=Item.objects.all())) \

    def get_worksheet_title(self, index=0):
        return gettext('Invoices')

    def get_queryset(self):
        return self.queryset


class InvoicePdfDetailExporter(ExporterMixin):
    queryset = Invoice.objects.all()
    export_format = Export.FORMAT_PDF
    export_context = Export.CONTEXT_DETAIL
    filename = _('invoices.zip')

    def get_queryset(self):
        return self.queryset

    def export(self):
        self.write_data(self.output)

    def write_data(self, output):
        export_files = get_invoices_in_pdf(self.get_queryset())

        if len(export_files) == 1:
            # directly export 1 PDF file
            file_data = export_files[0]
            self.filename = file_data['name']
            output.write(file_data['content'])
        else:
            # compress all invoices into single archive file
            output.write(compress(export_files).read())



# Generated by Django 2.0.3 on 2018-05-29 10:24

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('invoicing', '0005_auto_20171026_0009'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='invoice',
            options={
                'default_permissions': ('list', 'view', 'add', 'change', 'delete'),
                'ordering': ('date_issue', 'number'),
                'verbose_name': 'invoice',
                'verbose_name_plural': 'invoices'
            },
        ),
    ]



# Generated by Django 2.0.6 on 2018-06-05 10:32

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('invoicing', '0014_recalculate'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='invoice',
            options={'default_permissions': ('list', 'view', 'add', 'change', 'delete'), 'ordering': ('date_issue', 'sequence'), 'verbose_name': 'invoice', 'verbose_name_plural': 'invoices'},
        ),
    ]



from django.urls import re_path

from invoicing.views import InvoiceDetailView


app_name = 'invoicing'

urlpatterns = [
    re_path(r'^invoice/detail/(?P<pk>[-\d]+)/$', InvoiceDetailView.as_view(), name='invoice_detail'),
]



# Generated by Django 3.1.14 on 2023-02-07 19:29

from django.db import migrations
from internationalflavor.iban import BICField


class Migration(migrations.Migration):
    dependencies = [
        ('invoicing', '0026_auto_20210622_1055'),
    ]

    operations = [
        migrations.AlterField(
            model_name='invoice',
            name='bank_swift_bic',
            field=BICField(blank=True, max_length=11, verbose_name='Bank SWIFT / BIC'),
        ),
    ]



from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.http import HttpResponseForbidden
from django.shortcuts import get_object_or_404
from django.utils.decorators import method_decorator
from django.utils.module_loading import import_string
from django.views.generic import DetailView

from invoicing.models import Invoice


class InvoiceDetailView(DetailView):
    model = Invoice

    @method_decorator(login_required)
    def dispatch(self, request, *args, **kwargs):
        if not request.user.is_active and not request.user.is_superuser:
            return HttpResponseForbidden()

        invoice = get_object_or_404(self.model, pk=kwargs.get('pk', None))

        # TODO: refactor
        invoicing_formatter = getattr(settings, 'INVOICING_FORMATTER', 'invoicing.formatters.html.BootstrapHTMLFormatter')
        formatter_class = import_string(invoicing_formatter)
        formatter = formatter_class(invoice)
        return formatter.get_response()



# Generated by Django 2.1.10 on 2019-09-12 06:53

from django.db import migrations
import model_utils.fields


class Migration(migrations.Migration):

    dependencies = [
        ('invoicing', '0020_auto_20181001_1025'),
        ('invoicing', '0021_invoice_related_document'),
    ]

    operations = [
        migrations.AddField(
            model_name='invoice',
            name='date_paid',
            field=model_utils.fields.MonitorField(blank=True, default=None, monitor='status', null=True, when={'PAID'}),
        ),
    ]



from datetime import date
from decimal import Decimal
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.core.validators import EMPTY_VALUES
from internationalflavor.vat_number import VATNumberValidator

from invoicing.taxation import TaxationPolicy


class EUTaxationPolicy(TaxationPolicy):
    """
    This taxation policy should be correct for all EU countries. It uses following rules:
        * if supplier country is not in EU - assert error,
        * return 'default tax' in cases:
            * if supplier country and customer country are the same,
            * if supplier country and customer country are not the same, but customer is private person from EU,
            * if supplier country and customer country are not the same, customer is company from EU, but his VAT ID is not valid according VIES system.
        * return tax not applicable (None) in cases:
            * if supplier country and customer country are not the same, customer is company from EU and his tax id is valid according VIES system.
            * if supplier country and customer country are not the same and customer is private person not from EU,
            * if supplier country and customer country are not the same and customer is company not from EU.


    Please note, that term "private person" refers in system to user that did not provide tax ID and
    ``company`` refers to user that provides it.

    """
    EU_COUNTRIES_RATES = {
        'AT': Decimal(20),  # Austria
        'BE': Decimal(21),  # Belgium
        'BG': Decimal(20),  # Bulgaria
        'CY': Decimal(19),  # Cyprus
        'CZ': Decimal(21),  # Czech Republic
        'DK': Decimal(25),  # Denmark
        'EE': Decimal(22),  # Estonia
        'FI': Decimal(24),  # Finland
        'FR': Decimal(20),  # France
        'DE': Decimal(19),  # Germany
        'GR': Decimal(24),  # Greece
        'HR': Decimal(25),  # Croatia
        'HU': Decimal(27),  # Hungary
        'IE': Decimal(23),  # Ireland
        'IT': Decimal(22),  # Italy
        'LV': Decimal(21),  # Latvia
        'LT': Decimal(21),  # Lithuania
        'LU': Decimal(17),  # Luxembourg
        'MT': Decimal(18),  # Malta
        'NL': Decimal(21),  # Netherlands
        'PL': Decimal(23),  # Poland
        'PT': Decimal(23),  # Portugal
        'RO': Decimal(19),  # Romania
        'SK': [
            {"from": date.min, "to": date(2024, 12, 31), "rate": Decimal(20)},
            {"from": date(2025, 1, 1), "to": date.max, "rate": Decimal(23)},
        ], # Slovakia
        'SI': Decimal(22),  # Slovenia
        'ES': Decimal(21),  # Spain
        'SE': Decimal(25),  # Sweden
        # 'GB': Decimal(20),  # United Kingdom (Great Britain) - not in EU since 1.2.2020
    }

    @classmethod
    def is_in_EU(cls, country_code):
        return country_code.upper() in cls.EU_COUNTRIES_RATES.keys()

    @classmethod
    def is_reverse_charge(cls, invoice, delivery_country=None, check_items=True):
        supplier_country = invoice.supplier_country.code if invoice.supplier_country else None

        if not supplier_country:
            supplier_country = cls.get_supplier_country_code()

        # Supplier has to be from EU
        if not cls.is_in_EU(supplier_country):
            return False

        # Supplier VAT ID has to be set
        if invoice.supplier_vat_id in EMPTY_VALUES:
            return False

        # Customer VAT ID has to be set
        if invoice.customer_vat_id in EMPTY_VALUES:
            return False

        # Place of supply is either delivery country or customer country
        place_of_supply = delivery_country or invoice.customer_country

        # missing place of supply
        if place_of_supply in EMPTY_VALUES:
            return False

        # supplier and delivery countries have to be different
        if invoice.supplier_country == place_of_supply:
            return False

        # there has to be at least one invoice item with None tax rate
        if check_items and invoice.item_set.exists() and not invoice.item_set.filter(tax_rate=None).exists():
            return False

        # customer has to be from EU -> not True (for example Great Britain)
        # if not invoice.is_EU_customer():
        #     return False

        return True

    # TODO: rename to get_default_tax_rate
    @classmethod
    def get_default_tax(cls, country_code=None, tax_point_date=None):
        """
        Gets default tax rate.``

        :param country_code: The ISO country code
        :param tax_point_date: The date of the tax point (date object)
        :return: Decimal()
        """
        # Use the current time if tax_point_date is not provided
        if tax_point_date is None:
            tax_point_date = date.today()

        default_tax_rate = super().get_default_tax(country_code, tax_point_date)

        # If country code and tax point date are provided, fetch the rate
        if country_code and not hasattr(settings, 'INVOICING_TAX_RATE'):
            return cls.get_rate_for_country(country_code, tax_point_date)

        return default_tax_rate

    @classmethod
    def get_tax_rate(cls, invoice):
        if invoice.supplier_vat_id in EMPTY_VALUES:
            # Supplier is not a VAT payer
            return None

        # customer_country = invoice.customer_country.code if invoice.customer_country else None
        supplier_country = invoice.supplier_country.code if invoice.supplier_country else None

        if not supplier_country:
            supplier_country = cls.get_supplier_country_code()

        if not cls.is_in_EU(supplier_country):
            raise ImproperlyConfigured("EUTaxationPolicy requires that supplier country is in EU")

        # Reverse charge
        if cls.is_reverse_charge(invoice, check_items=False):
            if not getattr(settings, 'INVOICING_USE_VIES_VALIDATOR', True):
                return None

            try:
                # Verify VAT ID in VIES
                VATNumberValidator(eu_only=True, vies_check=True)(invoice.customer_vat_id)

                # Company is registered in VIES
                # Charge back
                return None
            except ValidationError:
                pass

        # return default tax
        return cls.get_default_tax(supplier_country, invoice.date_tax_point)

    @classmethod
    def get_rate_for_country(cls, country_code, tax_point_date):
        """
        Gets the tax rate for a specific country and date.

        :param country_code: ISO country code
        :param tax_point_date: Date for which the tax rate applies (date object)
        :return: Decimal
        """
        # Get default tax rate
        default_tax_rate = super().get_default_tax(country_code, tax_point_date)

        # Find rate definition for the country
        rates = cls.EU_COUNTRIES_RATES.get(country_code, default_tax_rate)

        # If rates are defined as a list (date-specific), determine the applicable rate
        if isinstance(rates, list):
            for rate_period in rates:
                if rate_period["from"] <= tax_point_date <= rate_period["to"]:
                    return rate_period["rate"]
            # Fallback if no date range matches (unlikely if ranges are defined properly)
            raise ValueError(f"No valid tax rate found for {country_code} on {tax_point_date}")

        # If rates are not date-specific, return the flat rate
        return rates


# -*- coding: utf-8 -*-
# Generated by Django 1.10.3 on 2017-05-22 10:07
from __future__ import unicode_literals

from django.db import migrations, models
import model_utils.fields


class Migration(migrations.Migration):

    dependencies = [
        ('invoicing', '0003_auto_20170123_1810'),
    ]

    operations = [
        migrations.AlterField(
            model_name='invoice',
            name='date_sent',
            field=model_utils.fields.MonitorField(blank=True, default=None, monitor='status', null=True, when=set(['SENT'])),
        ),
        migrations.AlterField(
            model_name='invoice',
            name='delivery_method',
            field=models.CharField(default='PERSONAL_PICKUP', max_length=64),
        ),
        migrations.AlterField(
            model_name='invoice',
            name='status',
            field=models.CharField(default='NEW', max_length=64),
        ),
        migrations.AlterField(
            model_name='invoice',
            name='type',
            field=models.CharField(default='INVOICE', max_length=64),
        ),
        migrations.AlterField(
            model_name='item',
            name='unit',
            field=models.CharField(default='PIECES', max_length=64),
        ),
    ]



# Generated by Django 2.0.3 on 2018-06-01 12:59

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('invoicing', '0010_recalculate_vat'),
    ]

    operations = [
        migrations.AlterField(
            model_name='item',
            name='title',
            field=models.TextField(),
        ),
    ]



# Generated by Django 4.0.8 on 2023-12-13 16:41

from django.apps import apps
from django.db import migrations, models


def unset_supplier_vat(*args, **kwargs):
    print("Removing supplier VAT ID from invoices where it shouldn't be set")
    Invoice = apps.get_model("invoicing", "Invoice")

    for invoice in Invoice.objects.exclude(supplier_vat_id=''):
        if not invoice.is_supplier_vat_id_visible():
            print(f'#{invoice.id}, number {invoice.number}, supplier VAT ID = {invoice.supplier_vat_id}')
            invoice.supplier_vat_id = ''
            invoice.save(update_fields=['supplier_vat_id'])


class Migration(migrations.Migration):

    dependencies = [
        ('invoicing', '0030_invoice_supplier_is_vat_payer'),
    ]

    operations = [
        migrations.RunPython(unset_supplier_vat, migrations.RunPython.noop)
    ]



import json

from django.conf import settings


LANGUAGES = getattr(settings, 'INVOICING_LANGUAGES', getattr(settings, 'LANGUAGES', []))
ACCOUNTING_SOFTWARE = getattr(settings, 'INVOICING_ACCOUNTING_SOFTWARE', None)
try:
    ACCOUNTING_SOFTWARE_API_DATA = json.loads(getattr(settings, 'INVOICING_ACCOUNTING_SOFTWARE_API_DATA', None))
except:
    ACCOUNTING_SOFTWARE_API_DATA = None
# ACCOUNTING_SOFTWARE_API_KEY = getattr(settings, 'INVOICING_ACCOUNTING_SOFTWARE_API_KEY', None)

# TODO: refactor
ACCOUNTING_SOFTWARE_IKROS_API_URL = getattr(settings, 'INVOICING_ACCOUNTING_SOFTWARE_IKROS_API_URL', 'https://eshops.inteo.sk/api/v1/invoices/')
ACCOUNTING_SOFTWARE_PROFIT365_API_URL = getattr(settings, 'INVOICING_ACCOUNTING_SOFTWARE_PROFIT365_API_URL', 'https://api.profit365.eu/1.6')
ACCOUNTING_SOFTWARE_MANAGER = getattr(settings, 'INVOICING_ACCOUNTING_SOFTWARE_MANAGER', None)
